"""
Motion tween equations.
Python port of Robert Penner's easing equations.

(c) 2003 Robert Penner, all rights reserved. 
(c) 2009 Alexandre Quessy

Since the original actionscript library license is BSD, I publish this Python port under 
the GNU General Public License version 3 or later.

They are best used with the time.time() function. 

Easing Equations v1.5
May 1, 2003
(c) 2003 Robert Penner, all rights reserved. 
This work is subject to the terms in 
http://www.robertpenner.com/easing_terms_of_use.html.  

These tweening defs provide different flavors of 
math-based motion under a consistent API. 

Types of easing:

   Linear
   Quadratic
   Cubic
   Quartic
   Quintic
   Sinusoidal
   Exponential
   Circular
   Elastic
   Back
   Bounce

Changes:
1.5 - added bounce easing
1.4 - added elastic and back easing
1.3 - tweaked the exponential easing defs to make endpoints exact
1.2 - inline optimizations (changing t and multiplying in one step)
      -- thanks to Tatsuo Kato for the idea

Discussed in Chapter 7 of 
Robert Penner's Programming Macromedia Flash MX
(including graphs of the easing equations)

http://www.robertpenner.com/profmx
http://www.amazon.com/exec/obidos/ASIN/0072223561/robertpennerc-20
"""
import math

#constants
LINEAR_TWEEN =  0
EASE_IN_QUAD = 1 
EASE_OUT_QUAD = 2
EASE_IN_SINE = 3
EASE_OUT_SINE = 4
EASE_IN_OUT_SINE = 5
EASE_IN_EXPO = 6
EASE_OUT_EXPO = 7

MAX_TWEEN_KIND_ID = 7 # so that we can use the % to switch 

# ----------------------------------------------
def linearTween(t, b, c, d):
    """
    simple linear tweening - no easing
    
    @param t: current time, 
    @param b: beginning value
    @param c: change in value
    @param d: duration
    
    The arguments are the same for all other easing functions.
    
    t and d can be in frames or seconds/milliseconds, or whatever unit.
    """
    return c*t/d + b

# QUADRATIC EASING: t^2 

# FIXME : seems buggy 
# (stays still for a frame at end.)
def easeInQuad(t, b, c, d):
    """
    quadratic easing in - accelerating from zero velocity
    """
    return c*(t/d)*t + b

# FIXME : seems buggy when some negative values
# (sways back & forth within positive numbers)
def easeOutQuad(t, b, c, d):
    """
    quadratic easing out - decelerating to zero velocity
    """
    return -c *(t/d)*(t-2) + b

# SINUSOIDAL EASING: sin(t)
def easeInSine(t, b, c, d):
    """
    sinusoidal easing in - accelerating from zero velocity
    """    
    return -c * math.cos(t/d * (math.pi/2)) + c + b

def easeOutSine(t, b, c, d):
    """
    sinusoidal easing out - decelerating to zero velocity
    """
    return c * math.sin(t/d * (math.pi/2)) + b

def easeInOutSine(t, b, c, d):
    """
    sinusoidal easing in/out - accelerating until halfway, then decelerating
    """
    return -c/2 * (math.cos(math.pi*t/d) - 1) + b

# EXPONENTIAL EASING: 2^t 
def easeInExpo(t, b, c, d):
    """
    exponential easing in - accelerating from zero velocity
    """
    if t==0:
        return b
    else:
        return c * math.pow(2, 10 * (t/d - 1)) + b

def easeOutExpo(t, b, c, d):
    """
    exponential easing out - decelerating to zero velocity
    """
    if t==d:
        return b+c
    else:
        return c * (-math.pow(2, -10 * t/d) + 1) + b

# ----------------------------------------------

kinds = {
    LINEAR_TWEEN :  linearTween,
    EASE_IN_QUAD :  easeInQuad,
    EASE_OUT_QUAD : easeOutQuad,
    EASE_IN_SINE :  easeInSine,
    EASE_OUT_SINE : easeOutSine,
    EASE_IN_OUT_SINE : easeInOutSine,
    EASE_IN_EXPO : easeInExpo,
    EASE_OUT_EXPO : easeOutExpo,
}


# ----------------------------------------------

class Tween(object):
    """
    Motion tween tool.
    
    @param start_val From value
    @param end_val To value
    @param duration How long it lasts. Can be any unit.
    @param current_time The current time. Must be same unit as duration.
    @param kind Integer to choose one of the possible tweening kinds.
    """
    def __init__(self, start_val, end_val, duration, current_time, kind=LINEAR_TWEEN):
        # TODO: use time from pygame or time ?
        # self.t = current_time # current time
        self.start_time = float(current_time)
        self.begin_val = start_val # beginning value
        self.change = end_val - start_val # change in value
        self.duration = duration # duration
        self.kind = kind
    
    def is_done(self, current_time):
        """
        Returns if the tweening is done or not.
        
        @return boolean Is done or not.
        
        Uses the >= comparison operator.
        """
        elapsed = float(current_time) - self.start_time
        return elapsed >= self.duration
        
    def tick(self, current_time):
        """
        Simple wrapper for easing methods.
        
        @param current_time The current time. Must be in the same unit as duration.
        @returns current value now.
        
        All the easing methods take those arguments :
        t: current time
        b: beginning value
        c: change in value
        d: duration
        """
        global kinds
        _method = linearTween
        try:
            _method = kinds[self.kind]
        except KeyError, e:
            raise Exception("Invalid tween type " + str(e.message))
        elapsed = float(current_time) - self.start_time
        if elapsed > self.duration:
            ret = self.begin_val + self.change
        else:
            ret = _method(elapsed, self.begin_val, self.change, self.duration)
            # we override the min and max functions according to direction.
            _min = min
            _max = max
            if self.change < 0.0: # invert min/max if negative
                _min = max
                _max = min
            ret = _max(self.begin_val, _min(self.begin_val + self.change, ret))
        # print "ret: ", ret, " time: ", current_time, " kind:", _method
        return ret

# ----------------------------------------------
if __name__ == '__main__':
    # simple test
    print "this test will tween from 0 to 100 in steps of 10"
    kind = LINEAR_TWEEN
    start_val = 0.0
    end_val = 100.0
    duration = 10.0
    start_time = 20.0
    tween = Tween(start_val, end_val, duration, start_time, kind)
    for t in range(12):
        print tween.tick(t + start_time)

    print "same, but with in_out_sine"
    kind = EASE_IN_OUT_SINE
    tween = Tween(start_val, end_val, duration, start_time, kind)
    for t in range(12):
        print tween.tick(t + start_time)

